@@ -7,26 +7,24 @@ from typing import Any, Dict, TYPE_CHECKING
 
 import numpy as np
 import pandas as pd
+
 import sqlalchemy as alch
 from sqlalchemy.orm import backref, relationship
-import pyodbc
+from sqlalchemy.engine.default import DefaultDialect
+from sqlalchemy.dialects import mssql
 
 from subtypes import Frame, Enum
-from pathmagic import File
-from miscutils import lazy_property
+from pathmagic import File, PathLike
+from miscutils import cached_property
 from iotools.misc.serializer import LostObject
 
-from .custom import ModelMeta, Model, AutoModel, Query, Session, ForeignKey, Relationship, StringLiteral, BitLiteral
-from .expression import Select, Update, Insert, Delete, SelectInto
-from .utils import StoredProcedure, Script
-from .log import SqlLog
-from .database import Database, OrmSchemas, ObjectSchemas, OrmSchema, ObjectSchema, Metadata
-from .config import Config, Url
+from sqlhandler.custom import ModelMeta, Model, AutoModel, ReflectedModel, Query, Session, ForeignKey, Relationship, SubtypesDateTime, SubtypesDate, BitLiteral, Select, Update, Insert, Delete, SelectInto
+from sqlhandler.utils import StoredProcedure, Script, SqlLog
+from sqlhandler.database import Database, Metadata, Schemas, Schema
+from sqlhandler.utils import Config, Url
 
 if TYPE_CHECKING:
-    import alembic
-
-assert pyodbc
+    from alembic.operations import Operations
 
 
 class Sql:
@@ -41,21 +39,22 @@ class Sql:
         FAIL, REPLACE, APPEND = "fail", "replace", "append"
 
     class Constructors:
-        ModelMeta, Model, AutoModel = ModelMeta, Model, AutoModel
-        Database, OrmSchemas, ObjectSchemas, OrmSchema, ObjectSchema, Metadata = Database, OrmSchemas, ObjectSchemas, OrmSchema, ObjectSchema, Metadata
+        ModelMeta, Model, AutoModel, ReflectedModel = ModelMeta, Model, AutoModel, ReflectedModel
+        Database, Metadata, Schemas, Schema = Database, Metadata, Schemas, Schema
         Config, Query, Session = Config, Query, Session
         StoredProcedure, Script = StoredProcedure, Script
         Config, Frame = Config, Frame
 
-    constructors = Constructors()
-
-    CACHE_METADATA = True
+    class Settings:
+        cache_metadata = reflect_tables = reflect_views = True
+        eager_reflection = False
+        lazy_schemas = {"information_schema"}
 
     def __init__(self, connection: str = None, database: str = None, log: File = None, autocommit: bool = False) -> None:
+        self.constructors, self.settings = self.Constructors(), self.Settings()
         self.config = self.constructors.Config()
 
         self.engine = self._create_engine(connection=connection, database=database)
-        self.engine.sql = self
         self.engine.connect()
 
         self.session = self.constructors.Session(bind=self.engine, query_cls=self.constructors.Query)
@@ -70,6 +69,7 @@ class Sql:
         self.AND, self.OR, self.CAST, self.CASE, self.TRUE, self.FALSE = alch.and_, alch.or_, alch.cast, alch.case, alch.true(), alch.false()
 
         self.Table, self.Column, self.ForeignKey, self.Index, self.CheckConstraint, self.Relationship, self.relationship, self.backref = alch.Table, alch.Column, ForeignKey, alch.Index, alch.CheckConstraint, Relationship, relationship, backref
+        self.Datetime, self.Date = SubtypesDateTime, SubtypesDate
         self.type, self.func, self.sqlalchemy = alch.types, alch.func, alch
 
     def __repr__(self) -> str:
@@ -101,19 +101,19 @@ class Sql:
         return self.database.auto_model
 
     @property
-    def orm(self) -> OrmSchemas:
+    def tables(self) -> Schemas:
         """Property controlling access to mapped models. Models will only appear for tables that have a primary key, and never for views. Schemas must be accessed before tables: E.g. Sql().orm.some_schema.some_table"""
-        return self.database.orm
+        return self.database.tables
 
     @property
-    def objects(self) -> ObjectSchemas:
+    def views(self) -> Schemas:
         """Property controlling access to raw database objects. Schemas must be accessed before tables: E.g. Sql().orm.some_schema.some_table"""
-        return self.database.objects
+        return self.database.views
 
-    @lazy_property
-    def operations(self) -> alembic.operations.Operations:
+    @cached_property
+    def operations(self) -> Operations:
         """Property controlling access to alembic operations."""
-        from alembic.migration import MigrationContext
+        from alembic.runtime.migration import MigrationContext
         from alembic.operations import Operations
 
         return Operations(MigrationContext.configure(self.engine.connect()))
@@ -125,11 +125,11 @@ class Sql:
 
     @log.setter
     def log(self, val: File) -> None:
-        self._log = SqlLog(path=val, active=False)
+        self._log = SqlLog(path=val)
 
-    def initialize_log(self, logname: str, logdir: str = None) -> SqlLog:
+    def initialize_log(self, name: str, location: PathLike = None) -> SqlLog:
         """Instantiates a matt.log.SqlLog object from a name and a dirpath, and binds it to this object's 'log' attribute. If 'active' argument is 'False', this method does nothing."""
-        self._log = SqlLog.from_details(log_name=logname, log_dir=logdir, active=False)
+        self._log = SqlLog.from_details(log_name=name, log_dir=location, active=False)
         return self._log
 
     # Conversion Methods
@@ -179,11 +179,11 @@ class Sql:
         if has_identity_pk:
             dtypes[primary_key] = alch.types.INT
 
-        dataframe.to_sql(engine=self.engine, name=table, if_exists=if_exists, index=False, index_label=None, primary_key=primary_key, schema=schema, dtype=dtypes)
+        dataframe.to_sql(engine=self.engine, name=table, if_exists=if_exists, index=False, primary_key=primary_key, schema=schema, dtype=dtypes)
 
-        table_object = self.orm[schema][table]
-        self.refresh_table(table=table_object)
-        return self.orm[schema][table]
+        model = self.tables[schema][table]()
+        self.database.refresh_table(table=model.__table__)
+        return self.tables[schema][table]()
 
     def orm_to_frame(self, orm_objects: Any) -> Frame:
         """Convert a homogeneous list of sqlalchemy.orm instance objects (or a single one) to a pandas DataFrame."""
@@ -202,35 +202,30 @@ class Sql:
 
     def _create_engine(self, connection: str, database: str) -> alch.engine.base.Engine:
         url = self._create_url(connection=connection, database=database)
-        return alch.create_engine(str(url), echo=False, dialect=self._create_literal_dialect(url.get_dialect()))
+
+        engine = alch.create_engine(str(url), echo=False, dialect=self._customize_dialect(url.get_dialect()()))
+        engine.sql = self
+
+        return engine
 
     def _create_url(self, connection: str, database: str) -> Url:
         return self.config.generate_url(connection=connection, database=database)
 
-    def _create_literal_dialect(self, dialect_class: alch.engine.default.DefaultDialect) -> alch.engine.default.DefaultDialect:
-        from sqlalchemy.dialects.mssql import dialect as mssql
-        from sqlalchemy.dialects.sqlite import dialect as sqlite
+    def _customize_dialect(self, dialect: DefaultDialect) -> DefaultDialect:
+        dialect.colspecs.update(
+            {
+                alch.types.DateTime: SubtypesDateTime,
+                alch.types.DATETIME: SubtypesDateTime,
+                alch.types.Date: SubtypesDate,
+                alch.types.DATE: SubtypesDate,
+            }
+        )
 
-        class LiteralDialect(dialect_class):
-            supports_multivalues_insert = True
+        if isinstance(dialect, mssql.dialect):
+            dialect.supports_multivalues_insert = True
+            dialect.colspecs.update({alch.dialects.mssql.BIT: BitLiteral})
 
-            def __init__(self, *args: Any, **kwargs: Any) -> None:
-                super().__init__(*args, **kwargs)
-                self.colspecs.update(
-                    {
-                        alch.sql.sqltypes.String: StringLiteral,
-                        alch.sql.sqltypes.DateTime: StringLiteral,
-                        alch.sql.sqltypes.DATETIME: StringLiteral,
-                        alch.sql.sqltypes.Date: StringLiteral,
-                        alch.sql.sqltypes.DATE: StringLiteral,
-                        alch.sql.sqltypes.NullType: StringLiteral,
-                    }
-                )
-
-                if dialect_class in (mssql, sqlite):
-                    self.colspecs.update({alch.dialects.mssql.BIT: BitLiteral})
-
-        return LiteralDialect()
+        return dialect
 
     @staticmethod
     def _sql_dtype_dict_from_frame(frame: Frame) -> Dict[str, Any]:
@@ -239,8 +234,7 @@ class Sql:
 
         def sqlalchemy_dtype_from_series(series: pd.code.series.Series) -> Any:
             if series.dtype.name in ["int64", "Int64"]:
-                nums = [num for num in series if not isnull(num)]
-                if not nums:
+                if not (nums := [num for num in series if not isnull(num)]):
                     return alch.types.Integer
                 else:
                     minimum, maximum = min(nums), max(nums)
